/*
 * Copyright 2007, TransGaming Technologies Inc.
 *
 * Main reference:
 *   http://www.microsoft.com/technet/sysinternals/information/IoCompletionPorts.mspx
 */

#include "config.h"

#include "winternl.h"
#include "wine/server.h"
#include "wine/debug.h"

WINE_DEFAULT_DEBUG_CHANNEL(iocomp);


/******************************************************************************
 *		NtCreateIoCompletion (ntdll.@)
 */
NTSTATUS WINAPI NtCreateIoCompletion
(PHANDLE pIoCompletionHandle, ACCESS_MASK DesiredAccess,
 POBJECT_ATTRIBUTES pObjectAttributes, ULONG NumberOfConcurrentThreads)
{
   NTSTATUS Ret;

   TRACE ("(%p, 0x%x, %p, %lu)\n",
          pIoCompletionHandle, DesiredAccess, pObjectAttributes,
          NumberOfConcurrentThreads);

   if (pObjectAttributes)
      FIXME ("pObjectAttributes are not supported\n");

   if (pIoCompletionHandle == NULL)
      return STATUS_INVALID_PARAMETER;

   if (NumberOfConcurrentThreads == 0)
   {
      FIXME ("Need to query number of processors\n");
      NumberOfConcurrentThreads = 2;
   }

   SERVER_START_REQ(create_io_completion)
   {
      req->num_threads = NumberOfConcurrentThreads;
      Ret = wine_server_call (req);
      *pIoCompletionHandle = reply->handle;
   }
   SERVER_END_REQ;

   TRACE ("Ret=0x%lx, *pIoCompletionHandle=0x%x\n", Ret,
          *pIoCompletionHandle);
   return Ret;
}


/******************************************************************************
 *		NtRemoveIoCompletion (ntdll.@)
 */
NTSTATUS WINAPI NtRemoveIoCompletion
(HANDLE IoCompletionHandle, PULONG_PTR ppCompletionKey,
 LPOVERLAPPED *lplpOverlapped, PIO_STATUS_BLOCK pIoStatusBlock,
 PLARGE_INTEGER pWaitTime)
{
   NTSTATUS Ret;
   int cookie;

   TRACE ("(0x%x, %p, %p, %p, %p)\n", IoCompletionHandle,
          ppCompletionKey, lplpOverlapped, pIoStatusBlock, pWaitTime);

   if (!ppCompletionKey || !lplpOverlapped || !pIoStatusBlock)
      return STATUS_INVALID_PARAMETER;

   /* Ask for a task to do. If one's given to us, run with it. Otherwise,
      wait for one to be given to us */
   SERVER_START_REQ(remove_io_completion)
   {
      req->handle = IoCompletionHandle;
      req->cookie = &cookie;

      if (pWaitTime)
      {
         req->sec = pWaitTime->QuadPart / 1000000;
         req->usec = pWaitTime->QuadPart % 1000000;
         req->select_flags = SELECT_TIMEOUT;
      }
      else
      {
         req->sec = 0;
         req->usec = 0;
         req->select_flags = 0;
      }

      Ret = wine_server_call (req);
      pIoStatusBlock->u.Status = Ret;
      if (Ret == STATUS_SUCCESS)
      {
         *ppCompletionKey = (ULONG_PTR)reply->completion_key;
         *lplpOverlapped = reply->overlapped;
         pIoStatusBlock->Information = reply->num_bytes;
      }
   }
   SERVER_END_REQ;

   /* This can timeout immediately if a timeout of 0 is specified and
      no tasks are available */
   if ((Ret == STATUS_SUCCESS) || (Ret == STATUS_TIMEOUT))
   {
      TRACE ("Ret=0x%lx Key=0x%lx Overlapped=%p NumBytes=%lu\n", Ret,
             *ppCompletionKey, *lplpOverlapped, pIoStatusBlock->Information);
      return Ret;
   }

   if (Ret == STATUS_PENDING)
      Ret = wait_server_reply (&cookie);

   /* If we were told to abandon our waiting, that means we've been
      given a task to do */
   if (Ret == STATUS_ABANDONED_WAIT_0)
   {
      SERVER_START_REQ(retrieve_assigned_io_completion)
      {
         req->handle = IoCompletionHandle;
         Ret = wine_server_call (req);
         if (Ret == STATUS_SUCCESS)
         {
            *ppCompletionKey = (ULONG_PTR)reply->completion_key;
            *lplpOverlapped = reply->overlapped;
            pIoStatusBlock->Information = reply->num_bytes;
         }
      }
      SERVER_END_REQ;
   }

   pIoStatusBlock->u.Status = Ret;

   TRACE ("Ret=0x%lx Key=0x%lx Overlapped=%p NumBytes=%lu\n", Ret,
          *ppCompletionKey, *lplpOverlapped, pIoStatusBlock->Information);
   return Ret;
}


/******************************************************************************
 *		NtSetIoCompletion (ntdll.@)
 */
NTSTATUS WINAPI NtSetIoCompletion
(HANDLE IoCompletionHandle, ULONG_PTR pCompletionKey,
 LPOVERLAPPED lpOverlapped, ULONG NumberOfBytesTransferred,
 ULONG NumberOfBytesToTransfer)
{
   NTSTATUS Ret;

   TRACE ("(0x%x, 0x%lx, %p, %lu, %lu)\n",
          IoCompletionHandle, pCompletionKey, lpOverlapped,
          NumberOfBytesTransferred, NumberOfBytesToTransfer);

   if (NumberOfBytesTransferred)
      WARN ("NumberOfBytesTransferred != 0 not handled\n");


   SERVER_START_REQ(set_io_completion)
   {
      req->handle = IoCompletionHandle;
      req->completion_key = (void *)pCompletionKey;
      req->overlapped = lpOverlapped;
      req->num_bytes = NumberOfBytesToTransfer;
      Ret = wine_server_call (req);
   }
   SERVER_END_REQ;

   TRACE ("Ret=0x%lx\n", Ret);
   return Ret;
}
