/*
 * Shared Memory WineServer stuff.
 *
 * Copyright (C) 2002,2004,2006-2007 TransGaming Technologies
 */
#include "config.h"

#include <stdlib.h>
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#ifdef HAVE_SYS_IPC_H
# include <sys/ipc.h>
#endif
#ifdef HAVE_SEMAPHORE_H
# include <semaphore.h>
#endif
#ifdef HAVE_SYS_SHM_H
# include <sys/shm.h>
#endif
#ifdef HAVE_SYS_SEM_H
# include <sys/sem.h>
#endif
#include <stdio.h>
#include <string.h>
#include <signal.h>
#ifdef HAVE_SYS_MMAN_H
# include <sys/mman.h>
#endif

#include "config.h"
#include "wine/port.h"

#include "server_memory.h"
#include "request.h"
#include "mem_defs.h"
#include "mem_globals.h"

/* command-line options */
int debug_level = 0;
int persistent_server = 0;
int wineserver_is_daemon = 1;

/* Set to true only in wineserver process */
int is_wineserver = 0; // PH: FIXME: Remove hack

/* PH: FIXME: EVIL HACK */
int is_select = 0; 

/* Data area pointer */
server_memory_area* server_memory_block = NULL;
static size_t allocated_shm_size = 0;

/* this is set by wineserver (the binary) if wine_main_preload_info has been
 * set by the preloader.
 * have to do this double thinger because this symble is exported by the so,
 * not the binary (which is where the preloader looks).
 */
static void* reserved_shm_addr = NULL;
static size_t reserved_shm_size = 0;
void set_shm_reserved(void *addr, const size_t size)
{
    reserved_shm_addr = addr;
    reserved_shm_size = size;
}


#define DEBUG_LEVEL WITH_EXTREME_DEBUG

/* PH: FIXME: Need to have a way to communicate to grow a shared memory area. Is it possible? */
#define initial_size ( 4*1024*1024 ) /* Not too big for now... */

#ifndef __APPLE__
  #define shared_memory_address 0x90000000 /* FIXME DH: to go away (can almost go now) */
#else
  /* On the Mac, the 0x90000000 address is reserved */
  #define shared_memory_address 0xF0000000 /* FIXME DH: to go away (can almost go now) */
#endif

#define PAGESIZE getpagesize()

#define USE_CSRI_MALLOC

#if defined (HAVE_SEMAPHORE_H) && defined (HAVE_SHM_OPEN)
#  define USE_POSIX_IPC
#endif

/* SYSV IPC is fundamentally broken on Mac OS X, at least as of 10.4 */
#if defined (HAVE_SEMOP) && defined (HAVE_SHMGET) && defined (HAVE_FTOK) && !defined(__APPLE__)
#  define USE_SYSV_IPC
#endif

#define CRITSECTION_SEMA
#if defined( CRITSECTION_SEMA )
#  define RECURIVE_CRITSECION_SEMA
#endif
/* #define MEMORY_PROTECTION */ /* Doesn't work right now */
/* #define HANDLE_SHM_CRASHES */ /* Doesn't work right now */


void protect_memory(void);
void unprotect_memory(void);

static pid_t signal_tid;


#if defined( HANDLE_SHM_CRASHES )
static sigset_t block_sigset;
static sigset_t orig_sigset;
static struct sigaction shm_action;
static struct sigaction shm_orig_sigsegv_action;


// PH: FIXME: Need to do some serious cleanup here. Integrate into signal_i386.c
//            or something.
typedef struct
{
    unsigned short sc_gs, __gsh;
    unsigned short sc_fs, __fsh;
    unsigned short sc_es, __esh;
    unsigned short sc_ds, __dsh;
    unsigned long sc_edi;
    unsigned long sc_esi;
    unsigned long sc_ebp;
    unsigned long sc_esp;
    unsigned long sc_ebx;
    unsigned long sc_edx;
    unsigned long sc_ecx;
    unsigned long sc_eax;
    unsigned long sc_trapno;
    unsigned long sc_err;
    unsigned long sc_eip;
    unsigned short sc_cs, __csh;
    unsigned long sc_eflags;
    unsigned long esp_at_signal;
    unsigned short sc_ss, __ssh;
    unsigned long i387;
    unsigned long oldmask;
    unsigned long cr2;
} SIGCONTEXT;


static void unexpected_shm_signal_handler(int sig)
{
    /* Sigaction is unfortunately per process, not per thread so we may have to do a manual
     * dispatch to the correct signal handler for the running thread.
     */
    if( signal_tid != wine_get_inprocess_tid() )
    {
        shm_orig_sigsegv_action.sa_handler( sig );
        return;
    }

    info_msg( ALWAYS, current, "Fatal Signal during tid %d shmwineserver execution!\n", wine_get_inprocess_tid() );

    release_shm_server_semaphore();

    raise( SIGSTOP );
    _exit(SIGSEGV);
}
#endif /* HANDLE_SHM_CRASHES */

void setup_shm_signal_handling(void)
{
#if defined( HANDLE_SHM_CRASHES )

    sigset_t sigset;

    sigemptyset( &block_sigset );
    sigaddset( &block_sigset, SIGCHLD );
    sigaddset( &block_sigset, SIGHUP );
    sigaddset( &block_sigset, SIGINT );
    sigaddset( &block_sigset, SIGQUIT );
    sigaddset( &block_sigset, SIGTERM );

    sigemptyset( &sigset );
    sigaddset( &sigset, SIGSEGV );
    shm_action.sa_mask = block_sigset;
#if defined( linux )
    shm_action.sa_flags = SA_RESTART;
#else
    shm_action.sa_flags = 0;
#endif
    shm_action.sa_handler = unexpected_shm_signal_handler;

    sigaction( SIGSEGV, &shm_action, &shm_orig_sigsegv_action );
#endif /* HANDLE_SHM_CRASHES */
}

void signal_enter_shm(void)
{
    if( is_wineserver ) return;

    signal_tid = wine_get_inprocess_tid();
}

void signal_leave_shm(void)
{
    if( is_wineserver ) return;

    signal_tid = 0; // PH: FIXME: What's an invalid tid?
}


/* Function pointer definitions for shm/sem callbacks */
typedef void (*sem_open_t) (BOOL Create);
typedef void (*sem_destroy_t) ();
typedef void (*sem_acquire_t) ();
typedef void (*sem_release_t) ();
typedef BOOL (*shm_open_t) (const char *pName);
typedef void (*shm_destroy_t) ();
typedef void* (*shm_map_t) (BOOL Create, void *Addr, size_t Len);

typedef struct {
   sem_open_t sem_open;
   sem_destroy_t sem_destroy;
   sem_acquire_t sem_acquire;
   sem_release_t sem_release;
   shm_open_t shm_open;
   shm_destroy_t shm_destroy;
   shm_map_t shm_map;
} shm_ops_t;

static shm_ops_t shm_ops;

typedef enum {SHM_USING_NONE, SHM_USING_POSIX, SHM_USING_SYSV} shm_using_t;
static shm_using_t shm_type = SHM_USING_NONE;

#ifdef USE_POSIX_IPC

static char  *shm_path;
static int    shm_fd = -1;
static sem_t *shm_sem = (sem_t *)SEM_FAILED;

static int create_shm_path (const char *pathname)
{
   size_t len, i;

   shm_path = strdup (pathname);
   if (!shm_path)
   {
      perror ("Unable to copy shm path");
      return 0;
   }

   /* For portability, need to ensure it starts with a / and doesn't
      contain any others */
   if (shm_path[0] != '/')
      shm_path[0] = '/';

   len = strlen (shm_path);
   for (i = 1; i < len; i++)
   {
      if (shm_path[i] == '/')
         shm_path[i] = '_';
   }
   
#ifdef __APPLE__
   /* Apple doesn't like the name to be very long! */
   shm_path[30] = 0;
#endif

   return 1;
}

static void posix_sem_destroy ()
{
   if (sem_close (shm_sem))
      perror ("Error closing semaphore");
   if (sem_unlink (shm_path))
      perror ("Error unlinking semaphore");
}

static void posix_sem_open (BOOL Create)
{
   /* Ensure semaphore doesn't already exist */
   if (Create)
      sem_unlink (shm_path);

   /* Create our semaphore */
   if (Create)
      shm_sem = sem_open (shm_path, O_CREAT | O_EXCL, 0600, 0);
   else
      shm_sem = sem_open (shm_path, 0);
   if (shm_sem == (sem_t *)SEM_FAILED)
      fatal_perror ("sem_open failed");
}

static void posix_sem_acquire ()
{
   while (sem_wait (shm_sem))
   {
      if (errno == EINTR)
         continue;

      fatal_perror ("sem_wait failed");
   }
}

static void posix_sem_release ()
{
   if (sem_post (shm_sem))
      fatal_perror ("sem_post failed");
}

static void posix_shm_destroy ()
{
   close (shm_fd);
   if (shm_unlink (shm_path))
      perror ("Failed to destory shared memory\n");
}

static BOOL posix_shm_create (const char *pName)
{
   if (!create_shm_path (pName))
   {
      perror ("Unable to copy shm path");
      return FALSE;
   }

   /* Ensure it doesn't already exist */
   shm_unlink (shm_path);

   if (debug_level > 2)
      fprintf (stderr,"shm_path: %s\n", shm_path);

   shm_fd = shm_open (shm_path, O_CREAT | O_EXCL | O_RDWR, S_IRWXU);
   if (shm_fd < 0)
   {
      perror ("Unable to create POSIX shm segment");
      return FALSE;
   }

   return TRUE;
}

static BOOL posix_shm_open (const char *pName)
{
   if (!create_shm_path (pName))
   {
      perror ("Unable to copy shm path");
      return FALSE;
   }

   shm_fd = shm_open (shm_path, O_RDWR, 0);
   if (shm_fd < 0)
   {
      perror ("Unable to open POSIX shm segment");
      return FALSE;
   }

   return TRUE;
}

static void *posix_shm_map (BOOL Create, void *Addr, size_t Len)
{
   if (Create)
   {
      /* Set the shm entry size */
      if (ftruncate (shm_fd, Len))
      {
         perror ("Unable to set POSIX shm segment size");
         return (void *)-1;
      }
   }

   return mmap (Addr, Len, PROT_READ | PROT_WRITE, MAP_SHARED, shm_fd, 0);
}

static void posix_initialize_ops ()
{
   shm_type = SHM_USING_POSIX;
   shm_ops.sem_open = posix_sem_open;
   shm_ops.sem_destroy = posix_sem_destroy;
   shm_ops.sem_acquire = posix_sem_acquire;
   shm_ops.sem_release = posix_sem_release;
   shm_ops.shm_open = posix_shm_open;
   shm_ops.shm_destroy = posix_shm_destroy;
   shm_ops.shm_map = posix_shm_map;
}

#endif /* USE_POSIX_IPC */

#ifdef USE_SYSV_IPC

static const int shared_memory_project_id = 0x54477758; /* "TGwX" in ascii. */
static int shm_sema_id = 0;
static int shmid_server;
static key_t shmkey;

#if defined( _SEM_SEMUN_UNDEFINED )
union semun {
   int val;
   struct semid_ds *buf;
   ushort * array;
};
#endif /* _SEM_SEMUN_UNDEFINED */

static void sysv_sem_destroy ()
{
   if (semctl (shm_sema_id, 0, IPC_RMID) == -1)
      perror ("Error closing semaphore");
}

static void sysv_sem_open (BOOL Create)
{
   union semun argument;

   argument.val = 0; /* Initially the semaphore is not available */

   if (Create)
      shm_sema_id = semget (shmkey, 1, IPC_CREAT | SHM_R | SHM_W);
   else
      shm_sema_id = semget (shmkey, 1, SHM_R | SHM_W);

   if (shm_sema_id == -1)
      fatal_perror ("semget fail");
   
   /* Set the semaphore as available to start with */
   if (Create)
   {
      if (semctl (shm_sema_id, 0, SETVAL, argument) < 0)
         fatal_perror ("Can't set value for new semaphore");
   }
}

static void sysv_sem_acquire ()
{
   static struct sembuf acquire_operation = {0, -1, 0};

   while (semop (shm_sema_id, &acquire_operation, 1) == -1)
   {
      if (errno == EINTR)
         continue;

      fatal_perror ("semop acquire failed");
   }
}

static void sysv_sem_release ()
{
   static struct sembuf release_operation = {0, 1, 0};

   while (semop (shm_sema_id, &release_operation, 1) == -1)
   {
      if (errno == EINTR)
         continue;

      fatal_perror ("semop release failed");
   }
}

static void sysv_shm_destroy ()
{
   if (shmctl (shmid_server, IPC_RMID, 0) == -1)
      perror ("Failed to destory shared memory\n");
}

static BOOL sysv_shm_create (const char *pName)
{
   shmkey = ftok (pName, shared_memory_project_id);
   if (shmkey == -1)
   {
      perror ("can't create key_t for SYSV shm");
      return FALSE;
   }

   if (debug_level > 2)
      fprintf (stderr, "Key is 0x%x\n", shmkey);

   return TRUE;
}

static void *sysv_shm_map (BOOL Create, void *Addr, size_t Len)
{
   int flags = SHM_R | SHM_W;

   if (Create)
      flags |= IPC_CREAT;

   shmid_server = shmget (shmkey, Len, flags);
   if (shmid_server == -1)
      return (void *)-1;

   return shmat (shmid_server, Addr, 0);
}

static void sysv_initialize_ops ()
{
   shm_type = SHM_USING_SYSV;
   shm_ops.sem_open = sysv_sem_open;
   shm_ops.sem_destroy = sysv_sem_destroy;
   shm_ops.sem_acquire = sysv_sem_acquire;
   shm_ops.sem_release = sysv_sem_release;
   shm_ops.shm_open = sysv_shm_create;
   shm_ops.shm_destroy = sysv_shm_destroy;
   shm_ops.shm_map = sysv_shm_map;
}

#endif /* USE_SYSV_IPC */

static void atexit_destroy_semaphore(void)
{
   if (debug_level)
      fprintf (stderr, "atexit destroy of semaphore\n");
   shm_ops.sem_destroy ();
}

static void create_server_semaphore ()
{
   shm_ops.sem_open (TRUE);
   atexit (atexit_destroy_semaphore);
}

static void get_server_semaphore ()
{
   shm_ops.sem_open (FALSE);
}

#if defined( CRITSECTION_SEMA ) && defined( __i386__ )
/* PH: FIXME: lifted, and not well, from dlls/ntdll/critsection.c */
inline static PVOID interlocked_cmpxchg( PVOID *dest, PVOID xchg, PVOID compare )
{
    PVOID ret;
    __asm__ __volatile__( "lock; cmpxchgl %2,(%1)"
                          : "=a" (ret) : "r" (dest), "r" (xchg), "0" (compare) : "memory" );
    return ret;
}
inline static LONG interlocked_inc( PLONG dest )
{
    LONG ret;
    __asm__ __volatile__( "lock; xaddl %0,(%1)"
                          : "=r" (ret) : "r" (dest), "0" (1) : "memory" );
    return ret + 1;
}

inline static LONG interlocked_dec( PLONG dest )
{
    LONG ret;
    __asm__ __volatile__( "lock; xaddl %0,(%1)"
                          : "=r" (ret) : "r" (dest), "0" (-1) : "memory" );
    return ret - 1;
}
#elif defined( CRITSECTION_SEMA ) && defined(__ppc__)
inline static PVOID interlocked_cmpxchg( PVOID *dest, PVOID xchg, PVOID compare )
{
#warning THIS is wrong, there is no exchange going on on the PPC FIXMEEEEEEEEEEE
    PVOID ret;
/*    __asm__ __volatile__( "lock; cmpxchgl %2,(%1)"
                          : "=a" (ret) : "r" (dest), "r" (xchg), "0" (compare) : "memory" );
*/    return ret;
}
inline static LONG interlocked_inc( PLONG dest )
{
#warning THIS is wrong, there is no exchange going on on the PPC FIXMEEEEEEEEEEE
    LONG ret;
/*    __asm__ __volatile__( "lock; xaddl %0,(%1)"
                          : "=r" (ret) : "r" (dest), "0" (1) : "memory" );
*/    return ret + 1;
}

inline static LONG interlocked_dec( PLONG dest )
{
#warning THIS is wrong, there is no exchange going on on the PPC FIXMEEEEEEEEEEE
    LONG ret;
/*    __asm__ __volatile__( "lock; xaddl %0,(%1)"
                          : "=r" (ret) : "r" (dest), "0" (-1) : "memory" );
*/    return ret - 1;
}


#elif defined( CRITSECTION_SEMA )
#error "Define functions for your platform"
#endif /* CRITSECTION_SEMA */

#if defined( RECURIVE_CRITSECION_SEMA )
static inline void initialize_recursion(void)
{
  SERVER_GLOBAL_VARIABLE( sema_owning_pid ) = getpid();
  SERVER_GLOBAL_VARIABLE( sema_owning_tid ) = wine_get_inprocess_tid(); /* We start locked for wineserver */
  SERVER_GLOBAL_VARIABLE( sema_recursion_count ) = 1; /* We start locked for wineserver */
}
#endif /* RECURIVE_CRITSECION_SEMA */

/* Called prior to performing anything on shared data */
void acquire_shm_server_semaphore_exclusive(void)
{
   pid_t pid,tid;

   /* Don't do anything if we're not running as a shared memory server */
   if( !SERVER_GLOBAL_VARIABLE( shared_memory_server ) )
     return;

   pid = getpid();
   tid = wine_get_inprocess_tid();

   if( debug_level > DEBUG_LEVEL ) fprintf( stderr, "%d - Attempting to get sema\n", tid );
     
#if defined( CRITSECTION_SEMA )
   if( interlocked_inc( &SERVER_GLOBAL_VARIABLE( sema_lock_count ) ))
   {
#endif

#if defined( RECURIVE_CRITSECION_SEMA )
      if( (pid == SERVER_GLOBAL_VARIABLE( sema_owning_pid )) &&
          (tid == SERVER_GLOBAL_VARIABLE( sema_owning_tid )))
     {
       SERVER_GLOBAL_VARIABLE( sema_recursion_count )++;
       return;
     }
     else
#endif
     {
       if (debug_level > DEBUG_LEVEL)
          fprintf (stderr, "%d - Taking slow path\n", tid);
       shm_ops.sem_acquire ();
     }
  
#if defined( CRITSECTION_SEMA ) 
   }
#endif

   SERVER_GLOBAL_VARIABLE( sema_owning_pid ) = pid;
   SERVER_GLOBAL_VARIABLE( sema_owning_tid ) = tid;
   SERVER_GLOBAL_VARIABLE( sema_recursion_count ) = 1;

   if( debug_level > DEBUG_LEVEL ) fprintf( stderr, "%d - Got sema\n", tid );
   
   signal_enter_shm();
   unprotect_memory();
}

/* Call to release */
void release_shm_server_semaphore(void)
{
   /* Don't do anything if we're not running as a shared memory server */
   if( !SERVER_GLOBAL_VARIABLE( shared_memory_server ) )
     return;

   protect_memory();
   signal_leave_shm();

#if defined( RECURIVE_CRITSECION_SEMA )
   if( --SERVER_GLOBAL_VARIABLE( sema_recursion_count ) )
   {
     interlocked_dec( &SERVER_GLOBAL_VARIABLE( sema_lock_count ) );
   }
   else
   {
     SERVER_GLOBAL_VARIABLE( sema_owning_tid ) = 0;
     SERVER_GLOBAL_VARIABLE( sema_owning_pid ) = 0;
#endif
#if defined( CRITSECTION_SEMA )
     if( interlocked_dec( &SERVER_GLOBAL_VARIABLE( sema_lock_count ) ) >= 0 )
     {
#endif
        shm_ops.sem_release ();
#if defined( CRITSECTION_SEMA )
     }
#endif

#if defined( RECURIVE_CRITSECION_SEMA )
   }
#endif /* RECURIVE_CRITSECION_SEMA */

   if( debug_level > DEBUG_LEVEL ) fprintf( stderr, "%d - Released sema\n", wine_get_inprocess_tid() );
}



static void atexit_destroy_shared_memory(void)
{
   if (debug_level)
      fprintf (stderr, "atexit destroy of shared_memory\n");
   shm_ops.shm_destroy ();
}


void* regular_memory_alloc( const char* pathname, int reg_size )
{
  /* Just use malloc since things aren't up and running yet */
  return malloc( initial_size + reg_size ); 
}

/* Anything which needs doing prior to using shm_mem_malloc */
static void setup_shm_malloc(void)
{
#if DEBUG
  _malloc_statsfile = stderr;
#endif /* DEBUG */
}

/* Setup shared memory for the master owner (wineserver process in this case) */
void* create_server_shared_memory_area( const char* pathname, int reg_size )
{
  void *shm_addr, *shm_get_addr;
  size_t shm_size = initial_size + reg_size;
  BOOL ShmInitialized = FALSE;

  /* If we've already established the segment for this process, there's not need to 
   * repeat.
   */
  if( server_memory_block )
    return server_memory_block;

  /* figure out the address we should use */
  shm_get_addr = (void*)shared_memory_address;
  if (reserved_shm_size)
  {
      shm_get_addr = reserved_shm_addr;
      /* now unmmap it, cause we are going to use it for shm */
      if (debug_level > 2)
          fprintf(stderr, "using reserved memory, munmap(%p, 0x%lx)\n",
                  shm_get_addr, (unsigned long)reserved_shm_size);
      munmap(shm_get_addr, reserved_shm_size);
  }
  if (reserved_shm_size && reserved_shm_size < shm_size)
  {
      fprintf(stderr, "sorry, but the preloader didn't reserve enough memory. "
              "preloader reserved 0x%08lx, we want 0x%08lx\n",
              (unsigned long)reserved_shm_size, (unsigned long)shm_size);
      return NULL;
  }

#ifdef USE_POSIX_IPC    
  ShmInitialized = posix_shm_create (pathname);
  if (ShmInitialized)
     posix_initialize_ops ();
#endif
#if defined(USE_SYSV_IPC)
  if (!ShmInitialized)
  {
     ShmInitialized = sysv_shm_create (pathname);
     if (ShmInitialized)
        sysv_initialize_ops ();
  }
#endif

  if (!ShmInitialized)
     return NULL;

  /* Try full size first; if no luck, keep trying smaller sizes until
     we succeed or hit 0 */
  while (shm_size > 0)
  {
     shm_addr = shm_ops.shm_map (TRUE, shm_get_addr, shm_size);
     if (shm_addr != (void *)-1)
        break;

     if (shm_size <= 2 * PAGESIZE)
        shm_size = 0;
     else
        shm_size -= 2 * PAGESIZE;
  }

  if (shm_size == 0)
  {
     fprintf (stderr,
              "Unable to allocate any shared memory. Is shm supported on your system?\n");
     shm_ops.shm_destroy ();
     return NULL;
  }
  else if (shm_size < (initial_size + reg_size))
  {
     if (shm_size < (reg_size + 0.5 * initial_size))
     {
        /* Probably not enough memory so don't use shm */
        fprintf (stderr,
                 "A less than optimial amount of shared memory was available (%lu vs %d).\n"
                 "Perhaps your registry is too large?\n",
                 (unsigned long)shm_size, initial_size + reg_size);
        shm_ops.shm_destroy ();
        return NULL;
     }

     if (debug_level)
        fprintf (stderr, "Using shm with less than ideal shm (%lu vs %d).\n",
                 (unsigned long)shm_size, initial_size + reg_size);
  }

  if (debug_level > 1)
     fprintf (stderr, "Allocated %lu with min_size %d and initial size %d\n",
              (unsigned long)shm_size, reg_size, initial_size);

  /* Setup for this process */
  server_memory_block = shm_addr;
  allocated_shm_size  = shm_size;

  /* Other processes cannot attach if we delete immediately. Mark for
     destruction on exit. */
  atexit (atexit_destroy_shared_memory);

  /* Initialize the memory */
  memset( server_memory_block, 0xff, allocated_shm_size );

  create_server_semaphore ();

  setup_shm_malloc();

  return shm_addr;
}

void* get_server_shared_memory_area(shm_using_t shm_server_type,
                                    const char* pathname, size_t shm_size,
                                    void *shm_get_addr, size_t shm_res_size,
                                    void *shm_res_addr )
{
  void* shm_addr;

  /* If we've already established the segment for this process, there's no
     need to repeat */
  if( server_memory_block )
    return server_memory_block;

#ifdef USE_POSIX_IPC
  if (shm_server_type == SHM_USING_POSIX)
     posix_initialize_ops ();
#endif
#ifdef USE_SYSV_IPC
  if (shm_server_type == SHM_USING_SYSV)
     sysv_initialize_ops ();
#endif
  if (shm_type == SHM_USING_NONE)
     return NULL;

  if (!shm_ops.shm_open (pathname))
     return NULL;

  /* if the preloader has reserved memory memory for us, free it here
   * just before the attach.*/
  if (shm_res_size &&
      ((shm_res_addr != shm_get_addr) || (shm_res_size < shm_size)))
     fprintf (stderr,
              "the memory reserved for the shm server by the preloader does\n"
              "not match the memory that the server is using. Will use the\n"
              "servers addr, but it may fail to allocate\n");

  if (shm_res_size)
  {
     if (debug_level > 2)
        fprintf (stderr, "Unmapping reserved area of %p (0x%lx)\n",
                 shm_res_addr, (unsigned long)shm_res_size);
     munmap (shm_res_addr, shm_res_size);
  }

  /* Get the segment */
  shm_addr = shm_ops.shm_map (FALSE, shm_get_addr, shm_size);
  if( shm_addr == (void*)-1 )
  {
    perror( "can't map memory (client)?" );
    return NULL;
  }

  /* Setup for this process */
  server_memory_block = shm_addr;
  allocated_shm_size  = shm_size;

  /* Mark the segment to be destroyed when not used anymore */
  /* PH: FIXME: I think that this needs to turn into an atexit( shmdt... */
  //shmctl( shmid, IPC_RMID, 0 );

  get_server_semaphore ();

  setup_shm_malloc();

  setup_shm_signal_handling();

  return shm_addr;
}




struct dynamic_memory_management
{
  size_t size;  /* Size of the dynamic memory area (constant) */
  void*  start; /* Start of the dynamic memory area (constant) */
  void*  free;  /* Ptr to the not yet used area */
};

#define DYN_MEM_AREA( smb )  ((struct dynamic_memory_management*)((smb)+1))

extern size_t shm_allocator_size(void);
extern void shm_setup_memory( void* ptr );

void init_memory_block( int shm )
{
  memset( server_memory_block, 0, sizeof( *server_memory_block ) );
  
  DYN_MEM_AREA(server_memory_block)->size = allocated_shm_size - sizeof( server_memory_block ) - sizeof(struct dynamic_memory_management) - shm_allocator_size();

  /* Make this page aligned for some malloc speed and to give a buffer before user data */
  DYN_MEM_AREA(server_memory_block)->free = (void*)((char*)server_memory_block + ((sizeof(*server_memory_block) + sizeof(struct dynamic_memory_management) + shm_allocator_size() + PAGESIZE-1) & ~(PAGESIZE-1)));
  DYN_MEM_AREA(server_memory_block)->start = DYN_MEM_AREA(server_memory_block)->free;
  
  /* Setup memory for the memory allocator variables */
  shm_setup_memory( (void*)((char*)server_memory_block + (sizeof(*server_memory_block) + sizeof(struct dynamic_memory_management)) ));

  SERVER_GLOBAL_VARIABLE( first_thread ) = NULL;
  SERVER_GLOBAL_VARIABLE( threadids ) = NULL;
  SERVER_GLOBAL_VARIABLE( first_free_threadid ) = NULL;
  SERVER_GLOBAL_VARIABLE( last_free_threadid ) = NULL;
#if 0
  /* FIXME MOOF NOTE: Currently the threads and processes won't
     share the same IDs. ie. they come from the same pool. */
  SERVER_GLOBAL_VARIABLE( processids ) = NULL;
  SERVER_GLOBAL_VARIABLE( first_free_processid ) = NULL;
  SERVER_GLOBAL_VARIABLE( last_free_processid ) = NULL;
#endif
  SERVER_GLOBAL_VARIABLE( server_start_time.tv_sec) = 0;
  SERVER_GLOBAL_VARIABLE( server_start_time.tv_usec) = 0;
  SERVER_GLOBAL_VARIABLE( shared_memory_server ) = shm;
  SERVER_GLOBAL_VARIABLE( shared_memory_server_size ) = allocated_shm_size;
  SERVER_GLOBAL_VARIABLE( shared_memory_server_addr ) = server_memory_block;
  SERVER_GLOBAL_VARIABLE( shared_memory_server_type ) = shm_type;
  SERVER_GLOBAL_VARIABLE( serverdir ) = NULL;
  SERVER_GLOBAL_VARIABLE( sema_lock_count ) = 0; /* We start locked for wineserver */
#if defined( RECURIVE_CRITSECION_SEMA )
  SERVER_GLOBAL_VARIABLE( sema_owning_pid ) = getpid();
  SERVER_GLOBAL_VARIABLE( sema_owning_tid ) = wine_get_inprocess_tid(); /* We start locked for wineserver */
  SERVER_GLOBAL_VARIABLE( sema_recursion_count ) = 1; /* We start locked for wineserver */
#endif

  /* NOTE: Can now use mem_alloc and other memory wrapping functions */

  STATIC_INITIALIZER_FUNCTION_CALL( atom );
  STATIC_INITIALIZER_FUNCTION_CALL( cdrom_eject );
  STATIC_INITIALIZER_FUNCTION_CALL( console );
  STATIC_INITIALIZER_FUNCTION_CALL( dbus );
  STATIC_INITIALIZER_FUNCTION_CALL( distributed );
  STATIC_INITIALIZER_FUNCTION_CALL( file );
  STATIC_INITIALIZER_FUNCTION_CALL( handle );
  STATIC_INITIALIZER_FUNCTION_CALL( mapping );
  STATIC_INITIALIZER_FUNCTION_CALL( object );
  STATIC_INITIALIZER_FUNCTION_CALL( process );
  STATIC_INITIALIZER_FUNCTION_CALL( registry );
  STATIC_INITIALIZER_FUNCTION_CALL( rpc_epmap );
  STATIC_INITIALIZER_FUNCTION_CALL( schedule );
  STATIC_INITIALIZER_FUNCTION_CALL( select );
  STATIC_INITIALIZER_FUNCTION_CALL( thread );
  STATIC_INITIALIZER_FUNCTION_CALL( user );
  STATIC_INITIALIZER_FUNCTION_CALL( window );
}


#if defined( USE_CSRI_MALLOC )

static void* simple_allocator(size_t size)
{
  void* new_area = DYN_MEM_AREA(server_memory_block)->free;
  
  if( (char*)DYN_MEM_AREA(server_memory_block)->free + size <= (char*)server_memory_block + DYN_MEM_AREA(server_memory_block)->size )
  {
    DYN_MEM_AREA(server_memory_block)->free = (char*)DYN_MEM_AREA(server_memory_block)->free + size;
  }
  else
  {
    /* Not enough memory available for the request */
    fatal_error( "not enough memory for allocation of %u bytes\n"
                 "start=%p present=%p total=0x%x\n",
                 (unsigned int)size,
                 DYN_MEM_AREA(server_memory_block)->start,
                 DYN_MEM_AREA(server_memory_block)->free,
                 (unsigned int)DYN_MEM_AREA(server_memory_block)->size );
    new_area = (void*)-1;
  }
  
  return new_area;
}

/* For hooking into the CSRI memory allocator */
void* (*_malloc_memfunc)(size_t) = simple_allocator;

#endif /* USE_CSRI_MALLOC */

#define CALLBACK_TAG( name ) get_##name##_callback
static const get_callback_function callback_lookup_table[ last_callback_tag_value - first_callback_tag_value + 1 ] =
{
  CALLBACK_TAG( queue_timer_timeout ),
  CALLBACK_TAG( timer_timer_timeout ),
  CALLBACK_TAG( queue_result_timeout ),
  CALLBACK_TAG( async ),
  CALLBACK_TAG( periodic_save ),
  CALLBACK_TAG( cdrom_eject ),
  CALLBACK_TAG( thread_thread_timeout ),
  CALLBACK_TAG( shuffle_threads ),
};
#undef CALLBACK_TAG

/* Virutalized multiprocess callbacks */
timeout_callback get_callback( timeout_callback cb )
{
  if( (long)cb < first_callback_tag_value || (long)cb > last_callback_tag_value )
  {
    fprintf( stderr, "Ah. Unimplemented callback %p\n", cb );
    *(int*)0 = 0x89;
  }

  return callback_lookup_table[ (long)cb - (long)first_callback_tag_value ]();
}


/* The order must be correct or obviously the mapping doesn't hold up */
#define OPS_TAG( name ) get_this_process_##name
static const ops_function ops_lookup_table[ last_ops_tag_value - first_ops_tag_value + 1 ] =
{
  OPS_TAG( atom_table_ops ),
  OPS_TAG( change_ops ),
  OPS_TAG( console_input_ops ),
  OPS_TAG( console_input_events_ops ),
  OPS_TAG( screen_buffer_ops ),
  OPS_TAG( debug_event_ops ),
  OPS_TAG( debug_ctx_ops ),
  OPS_TAG( device_ops ),
  OPS_TAG( event_ops ),
  OPS_TAG( file_ops  ),
  NULL,
  OPS_TAG( handle_table_ops ),
  OPS_TAG( mapping_ops ),
  OPS_TAG( mutex_ops ),
  OPS_TAG( named_pipe_ops ),
  OPS_TAG( pipe_user_ops ),
  OPS_TAG( process_ops ),
  OPS_TAG( startup_info_ops ),
  OPS_TAG( msg_queue_ops ),
  OPS_TAG( key_ops ),
  OPS_TAG( master_socket_ops ),
  OPS_TAG( semaphore_ops ),
  OPS_TAG( serial_ops ),
  OPS_TAG( snapshot_ops ),
  OPS_TAG( sock_ops ),
  OPS_TAG( thread_ops ),
  OPS_TAG( timer_ops ),
  OPS_TAG( fd_server_ops ),
  OPS_TAG( distributed_commands_ops ),
  OPS_TAG( dbus_table_ops ),
  OPS_TAG( iocomp_ops ),
  OPS_TAG( token_ops )
};

#undef OPS_TAG

/* Virtualized multiprocess object operations vtable */
const struct object_ops* get_ops( const struct object_ops* ops )
{
  if( (long)ops < first_ops_tag_value || (long)ops > last_ops_tag_value )
  {
    fprintf( stderr, "Ah. Unimplemented op %p\n", ops );
    *(int*)0 = 0x89;
  }

  return ops_lookup_table[ (long)ops - (long)first_ops_tag_value ]();
}


/* Entry point for freeing and allocating shared memory. Required if 1 side of library or other needs to allocate 
 * free and the other does the opposite.
 */
void* alloc_server_memory( size_t size )
{
  void* ptr;
  acquire_shm_server_semaphore_exclusive();
    ptr = mem_alloc( size );
  release_shm_server_semaphore();
  return ptr;
}

void free_server_memory( void* ptr )
{
  acquire_shm_server_semaphore_exclusive();
    mem_free( ptr );
  release_shm_server_semaphore();
  
}

void protect_memory(void)
{
#if defined( MEMORY_PROTECTION )
  int err;

  /* PH: FIXME: Kludge because the wineserver poll is broken */
  if( is_wineserver ) return;

  /* Cheat. We need access to the first bit of memory as that's where the semaphore stuff is */
  err = mprotect( (void*)server_memory_block+PAGESIZE, allocated_shm_size-PAGESIZE, PROT_NONE );

  if( err == -1 )
  {
    fprintf( stderr, "Protect failed: %d, %s\n", errno, strerror( errno ) );
  }
#endif
}

void unprotect_memory(void)
{
#if defined( MEMORY_PROTECTION )
  int err;

  /* PH: FIXME: Kludge because the wineserver poll is broken */
  if( is_wineserver ) return;

  /* Cheat. We need access to the first bit of memory as that's where the semaphore stuff is */
  err = mprotect( (void*)server_memory_block+PAGESIZE, allocated_shm_size-PAGESIZE, PROT_READ | PROT_WRITE );

  if( err == -1 )
  {
    fprintf( stderr, "Protect failed: %d, %s\n", errno, strerror( errno ) );
  }
#endif
}

void set_debug_level( int debuglevel )
{
  if( debuglevel || debug_level ) fprintf( stderr, "Setting the debug level to %d\n", debuglevel );
  debug_level = debuglevel;
}

/* Return the protocol version that is used for communications */
int shared_memory_protocol_version( void )
{
  return 1;
}

